package provider

import scala.concurrent.Future
import scala.concurrent.ExecutionContext.Implicits.global

import provider.ProviderUtils.firstSuccessOf

trait SearchQuery[T]:
  def matches(t: T): Boolean

case class MusicSearchQuery(
    artist: Option[String],
    album: Option[String],
    title: Option[String],
    year: Option[Int]
) extends SearchQuery[MusicEntry]:
  def matches(song: MusicEntry) =
    artist.forall(_ == song.artist) &&
      album.forall(_ == song.album) &&
      title.forall(_ == song.title) &&
      year.forall(_ == song.year)

class SearchProviderSync[Key, Entry](p: RacingProviderSync[Key, Entry])
    extends ProviderSync[SearchQuery[Entry], Set[Entry]]:
  def retrieveSync(query: SearchQuery[Entry]): Option[Set[Entry]] =
    val keys: Set[Key] =
      p.providerEU.retrieveKeysSync ++ p.providerUS.retrieveKeysSync
    Some((for
      k <- keys
      item <- p.providerEU.retrieveSync(k) ++ p.providerUS.retrieveSync(k)
      if query.matches(item)
    yield item).toSet)

  def retrieveKeysSync: Set[SearchQuery[Entry]] =
    throw UnsupportedOperationException(
      "Cannot retrieve keys of search provider"
    )

class SearchProvider[Key, Entry](p: RacingProvider[Key, Entry])
    extends Provider[SearchQuery[Entry], Set[Entry]]:
  def retrieve(query: SearchQuery[Entry]): Future[Set[Entry]] =
    def retrieveMatching(provider: Provider[Key, Entry]) =
      for
        keys <- provider.retrieveKeys
        items <- Future.traverse(keys)(p.retrieve)
      yield items.filter(query.matches)

    val fEU = retrieveMatching(p.providerEU)
    val fUS = retrieveMatching(p.providerUS)
    fEU.zip(fUS).map(_ ++ _)

    // Makes the last test fail:
    // val allKeys = p.providerEU.retrieveKeys.zip(p.providerUS.retrieveKeys).map(_ ++ _)
    // for
    //   keys <- allKeys
    //   items <- Future.traverse(keys)(p.retrieve)
    // yield items.filter(query.matches).toSet

  def retrieveKeys: Future[Set[SearchQuery[Entry]]] =
    throw UnsupportedOperationException(
      "Cannot retrieve keys of search provider"
    )
